#!/usr/bin/env python
import sys
import os
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
from csv import DictReader
from copy import copy
from collections import defaultdict

# LINE_STYLE = ['k-.', 'b--o', 'c--^', 'r--+', 'b:o', 'c:^', 'r:+', 'g-']
LINE_STYLE = ['k-.', 'g-', 'b--o', 'b:o', 'c--^', 'c:^', 'r--s', 'r:s']

# schednames = ['EDF', 'Naive-STP', 'PMCA-STP', 'PCWA-STP', 'Naive-PMTR', 'PMCA-PMTR', 'PCWA-PMTR', '(Naive+PMCA+PCWA)-(STP+PMTR)']
# legendnames= ['EDF', 'N/PST',     'CBA/PST',  'WBA/PST',  'N/RST',      'CBA/RST',   'WBA/RST',   'ALL']


schednames = ['EDF', '(Naive+PMCA+PCWA)-(STP+PMTR)', 'Naive-STP', 'Naive-PMTR', 'PMCA-STP', 'PMCA-PMTR', 'PCWA-STP', 'PCWA-PMTR']
legendnames= ['EDF',  'ALL'                        , 'N/PST'    , 'N/RST'     , 'CBA/PST' , 'CBA/RST'  ,  'WBA/PST',   'WBA/RST']


def main():
    datadir = "../results/"
    
    total_count = 0
#     'medium-medium-medium--harmonic-long--uni-heavy--uni-light--uni-heavy--way-heavy--uni-large--uni-heavy', 
#     filenames = ['light-light-heavy--harmonic-long--uni-heavy--uni-light--uni-light--regular--uni-large--uni-heavy', 'medium-medium-medium--harmonic-long--uni-light--uni-heavy--uni-light--regular--uni-small--uni-heavy']
#     for filename in filenames:
    for filename in os.listdir(os.getcwd() + "/" + datadir):
        d = DictReader(open(datadir + filename, 'r'))
        data = defaultdict(list)
    
        for row in d:
            for key, value in row.iteritems():
                data[key].append(value)
    
    
        npdata = defaultdict(list)
        npdata["sys_util"] = np.asarray(data["sys_util"])
        idx_ord = npdata["sys_util"].argsort()

        for key in data.keys():
            npdata[key] = np.asarray(data[key])[idx_ord]

        plt.figure(figsize=(8,4))
        plt.ylabel("Schedulability")
        plt.xlabel("Nominal System Utilization")
        for ndx, sched in enumerate(schednames):
            plt.plot(npdata["sys_util"], npdata[sched], LINE_STYLE[ndx], label=legendnames[ndx], linewidth=2.0)
            plt.xlim(0,9)
            
#         if filename is 'light-light-heavy--harmonic-long--uni-heavy--uni-light--uni-light--regular--uni-large--uni-heavy':
#             for ndx in range(len(npdata['Naive-STP'])):
#                 if float(npdata['Naive-STP'][ndx]) < 0.7 and float(npdata['Naive-STP'][ndx]) > 0.3:
#                     print "circling " + npdata["sys_util"][ndx] + ", " + npdata['Naive-STP'][ndx]
#                     plt.plot([npdata["sys_util"][ndx]], [npdata['Naive-STP'][ndx]], marker='o', markerfacecolor='none', markeredgecolor='r', ms=20.0, mew = 2.0)
#                     break
#                     
#         #         if npdata['util_dist'][0] == 'light-heavy-light' and npdata['task_util'][0] == 'uni-medium':
#         #              plt.plot([npdata["sys_util"][22]], [npdata['MC2-Iso'][22]], marker='s', markerfacecolor='none', markeredgecolor='g', ms=20.0, mew = 2.0)
#         # 
#         #         if npdata['util_dist'][0] == 'light-light-heavy':
#         #              plt.plot([npdata["sys_util"][35]], [npdata['MC2-Iso'][35]], marker='s', markerfacecolor='none', markeredgecolor='g', ms=20.0, mew = 2.0)
# 
#         
#         #         plt.legend(loc="upper right")
        plt.savefig("../plots/" + filename + ".pdf", bbox_inches='tight', pad_inches=0.1)

        plt.close()

        for count in npdata['Count']:
            if count != None:
                total_count += int(count)

#         break
#         fig = plt.figure(figsize = (1.5,1.3))
#         ax = plt.gca()
#         for style, sched in enumerate(legendnames):
#             plt.plot([], [], LINE_STYLE[style], label=sched+' ['+str(style+1)+']', linewidth=2.0)
#     
#         figleg = plt.figure(figsize = (8, 0.8))
#         leg = plt.figlegend(*ax.get_legend_handles_labels(), loc = 'upper left', ncol=4)
# 
#         leg.get_frame().set_edgecolor('w')
# 
#         figleg.savefig("./" + "sched-legend" + ".pdf")
#         break
#     
    print "count is " + str(total_count)

if __name__ == '__main__':
    main()