from __future__ import division

from schedcat.util.storage import storage

class TupleIndexer:
    """ A helper class for indexing design points by their id. """
    def __init__(self, divisor, mod):
        self.divisor = divisor
        self.mod = mod
    
    def index(self, i):
        return int(i / self.divisor) % self.mod

class DesignPointGenerator(object):

    def __init__(self, fal, is_valid = lambda dp: True):
        """
        fal is the factors and levels dictionary.

        """

        self.points = []
        self.fal = fal
        self.tuplecount = 1
        self.is_valid = is_valid
        self.master_index = 0
        divisor = 1
        self.indexers = {}

        for factor in sorted(self.fal.keys()):
            size = len(self.fal[factor])
            self.indexers[factor] = TupleIndexer(divisor, size)
            divisor = divisor * size
            self.tuplecount = self.tuplecount * size

    def __iter__(self):
        """ Make this class an iterator. """
        return self

    def next(self):
        """ Get the next design point in the iterator. """
        while True: # necessary if self.master_index is invalid
            if self.master_index == self.tuplecount:
                raise StopIteration
            else:
                point = storage()
                for factor in sorted(self.fal.keys()):
                    index = self.indexers[factor].index(self.master_index)
                    point[factor] = self.fal[factor][index]
                self.master_index += 1
                if self.is_valid(point):
                    return point
