#!/usr/bin/env python

from __future__ import division

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.']

from math   import ceil
import traceback
import argparse
import random
import time
import numpy as np
from collections import defaultdict
from csv import DictWriter
from util.stats import proportion_ci

from multiprocessing import Pool, cpu_count
from schedcat.util.storage import storage
from generator.generator import DesignPointGenerator

from distributions import *

from schedcat.model.LLC_params import *
from schedcat.generator.mcx_tasks import TaskGeneratorMCX
from LLCOptimizer import *
from LLCCanvas import LLCCanvas
from DAGGenerator import *
import schedcat.model.ohc as ohc
import loadohs
import graph

from sched_check import *

import multiprocessing
# We must import this explicitly, it is not imported by the top-level
# multiprocessing module.
import multiprocessing.pool
from random import randint

MIN_SAMPLES = 100
MAX_SAMPLES = 300 #2000
MAX_CI      = 0.05 #.05
CONFIDENCE  = 0.95 #.95

topdir = "~/LLC_Code/MCXSchedStudy/"

def make_mcx_task_set(dp):
    tgMCX = TaskGeneratorMCX(period = NAMED_PERIODS[dp.period],
                             util = NAMED_UTILIZATIONS[dp.task_util],
                             util_dist = NAMED_UTIL_DISTS[dp.util_dist],
                             mem_util = NAMED_MEMORY[dp.memory_util],
                             avg_percent = AVG_PERCENT,
                             a_inflation = uniform(1.5, 1.5),
                             mode_count = NAMED_MODE_COUNT[dp.mode_count],
                             mode_olap = NAMED_MODE_OVERLAP[dp.mode_olap],
                             crit_change = NAMED_CRIT_CHANGE[dp.crit_change],
                             util_var = NAMED_UTIL_VAR[dp.util_var],
                             pet_sens = NAMED_PET_SENS[dp.pet_sens],
                             dram_dist = NAMED_MEM_FOOTPRINTS[dp.dram],
                             stat_dist = NAMED_STATE_SIZE[dp.state_size],
                             pfuncs = P)
                             
    ts = tgMCX.make_task_set(max_util = dp.sys_util)
        
    ts.m = CORES
        
    ts.scheme = bypass_scheme()
    
    ts.modes = []
    for i in range(ts.mode_count):
        ts.modes.append(ts.mode(i))
        ts.modes[i].A = ts.modes[i].level('A')
        ts.modes[i].B = ts.modes[i].level('B')
        ts.modes[i].C = ts.modes[i].level('C')
        
    ts.dj = ts.disjoint()
    ts.dj.modes = []
    for i in range(ts.mode_count):
        ts.dj.modes.append(ts.dj.mode(i))
        ts.dj.modes[i].A = ts.dj.modes[i].level('A')
        ts.dj.modes[i].B = ts.dj.modes[i].level('B')
        ts.dj.modes[i].C = ts.dj.modes[i].level('C')
        
    ts.sh = ts.shared()
    for i in range(ts.mode_count):
        ts.sh.A = ts.sh.level('A')
        ts.sh.B = ts.sh.level('B')
        ts.sh.C = ts.sh.level('C')
    
    return ts

def complete(results, n):
    if n < MIN_SAMPLES:
        return False
    elif n > MAX_SAMPLES:
        return True
    else:
        for key in results:
            if proportion_ci(results[key], n, CONFIDENCE) > MAX_CI:
                return False
        return True

def update_mean2(old_mean, n, new_sample):
    return (old_mean*n + new_sample)/(n+1)

core_pgs_AB = 0
pgs_C = 0
bank_pgs = 128*256

mode_change_pgs_bank = 167
os_pgs_bank = 445

def set_mem_Naive(dp):
    global core_pgs_AB
    global pgs_C
    global os_pgs_bank
    core_pgs_AB = bank_pgs*0.25
    pgs_C = 2.25*bank_pgs
    os_pgs_bank = 160

def set_mem_WBA(dp):
    global core_pgs_AB
    global pgs_C
    global os_pgs_bank
    core_pgs_AB = bank_pgs
    os_pgs_bank = 445
    pgs_C = 2.25*bank_pgs

def mem_overflow_AB(ts):
    ts_AB = ts.level('A')
    ts_AB.extend(ts.level('B'))
    
#     print "   AB:"
    for core in range(0, 4):
#         print "    ", ts_AB.core(core).num_pages(), "/", core_pgs_AB-mode_change_pgs_bank-os_pgs_bank
        if ts_AB.core(core).num_pages() > core_pgs_AB-mode_change_pgs_bank-os_pgs_bank:
            return True
    
    return False
    
# def mem_overflow_AB_CBA_PST(ts):
#     ts_AB = ts.level('A')
#     ts_AB.extend(ts.level('B'))
#     
#     if ts_AB.core(0).num_pages() > core_pgs_AB-mode_change_pgs_bank-os_pgs_bank:
#         return True
#         
#     
#     for b_i in range(1, 4):
#         for g_0 in range(0, 3): #g_0: first mode we start with (not the 0th indexed mode)
#             sec_pp = np.sum([ts_AB.core(b_i).mode(g_i).num_pages() for g_i in range(g_0, ts.mode_count, 3)])
#             if sec_pp > core_pgs_AB-mode_change_pgs_bank-os_pgs_bank:
#                 return True
#     
#     return False
#     
# def mem_overflow_AB_CBA_RST(ts):
#     ts_AB = ts.level('A')
#     ts_AB.extend(ts.level('B'))
#         
#     for b_i in range(0, 4):
#         for g_0 in range(0, 4):
#             sec_pp = np.sum([ts_AB.core(b_i).mode(g_i).num_pages() for g_i in range(g_0, ts.mode_count, 4)])
#             if sec_pp > core_pgs_AB-mode_change_pgs_bank-os_pgs_bank:
#                 return True
#     
#     return False
    
def mem_ps_AB_CBA_PST(ts):
    ts_AB = ts.level('A')
    ts_AB.extend(ts.level('B'))
        
    mc_pgs_col = 11
    mc_pgs_os = 40
    ppm = 2048-mc_pgs_col-mc_pgs_os
    
    prts = [0, 0, 0, 0]
    
    prts[0] = int(ceil(ts_AB.core(0).num_pages()/ppm))
    
#     print "   AB:"
    for g_0 in range(0, 3):
        for b_i in range(1, 4):
            sec_pp = np.sum([ts_AB.core(b_i).mode(g_i).num_pages() for g_i in range(g_0, ts.mode_count, 3)])
            
        prts[g_0+1] = int(max(prts[g_0+1], ceil(sec_pp/ppm)))
    
    while np.sum(prts) < 16:
        val, idx = min((val, idx) for (idx, val) in enumerate(prts))
        prts[idx] += 1
    
#     print "    ", prts
    if np.sum(prts) <= 16:
        ABps = [area(0, prts[b_i]) for b_i in range(4)]
        Cps = area(0, 16)
        return scheme(ABps, ABps, Cps, 16, 16)
    return None
    
def mem_ps_AB_CBA_RST(ts):
    ts_AB = ts.level('A')
    ts_AB.extend(ts.level('B'))
        
    mc_pgs_col = 11
    mc_pgs_os = 40
    ppm = 2048-mc_pgs_col-mc_pgs_os
    
    prts = [0, 0, 0, 0]
            
#     print "  CBA:"
    for g_0 in range(0, 4):
        for b_i in range(0, 4):
            sec_pp = np.sum([ts_AB.core(b_i).mode(g_i).num_pages() for g_i in range(g_0, ts.mode_count, 4)])
            prts[(g_0+b_i)%4] = int(max(prts[(g_0+b_i)%4], ceil(sec_pp/ppm)))
    
    while np.sum(prts) < 16:
        val, idx = min((val, idx) for (idx, val) in enumerate(prts))
        prts[idx] += 1
        
#     print "   ", prts
    if np.sum(prts) == 16:
        ABps = [area(0, prts[b_i]) for b_i in range(4)]
        Cps = area(0, 16)
        return scheme(ABps, ABps, Cps, 16, 16)
        
    return None
    
def mem_overflow_C(ts):
#     print "  C:"
#     print "   ", (ts.level('C').num_pages() + ts.bp), "/", pgs_C
    return ts.level('C').num_pages() + ts.bp > pgs_C
    
def psForMode(g_i, lpDRAM, usePST):
    C = area(0, 16)
    AB = []
    if usePST:
        AB.append(area(0, lpDRAM[len(lpDRAM)-1]))
        AB.extend([area(0, lpDRAM[g_i][b_j]) for b_j in range(0, 3)])
    else:
        AB = [area(0, lpDRAM[g_i][b_j]) for b_j in range(0, 4)]
        
    return scheme(AB, AB, C, 16, 16)
    
def process_dp(dp):
    
    time_res = defaultdict(float)
    sched_res = defaultdict(float)
    count = defaultdict(int)
    
    n = 0
    
    while not complete(sched_res, n):
        ts = make_mcx_task_set(dp)
        
        unmnwts = []
        loadwtsAB = []
        loadwtsC = []
    
        for tsk in ts:
            unmnwts.append(unmanwt())
            loadwtsAB.append(loadedinfAB())
        
        for tsk in ts.level('C'):
            loadwtsC.append(loadedinfC())
    
        ts.useCshared()
        ts.setLLCAlloc(bypass_scheme())
        ohc.SetOheads(ts, 'mc2') 
    
        #---------inflate for loads in MC-Cshared------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                    
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
    
        ts.add_buffer_copy_times()
        
#         print "C pages:", ts.level('C').num_pages(), "from", len(ts.level('C')), "tasks"
#         for tsk in ts.level('C'):
#             print " ", tsk.pt
#     
#         print "Naive-PST:"
        #-------------------Naive-PST----------------------#
        schedNaive_PST = False
        timeNaive_PST = time.clock()
        ts.coreAssignPST(0, 4)
    
        set_mem_Naive(dp)
        
        memNaive_PST = not (mem_overflow_AB(ts) or mem_overflow_C(ts))
        
        ts.setLLCAlloc(bypass_scheme())
        
        if memNaive_PST:
            schedNaive_PST = True
            for tsm in ts.modes:
                psNaive_PST, filler = LPLLCWays(tsm, 'Integer', csch = default_scheme(), olap = True, debug = False)
                if psNaive_PST == None:
                    psNaive_PST, filler = LPLLCWays(tsm, 'Integer', csch = default_scheme(), olap = False, debug = False)
                if psNaive_PST == None:
                    schedNaive_PST = False
                    break
                    
        timeNaive_PST = time.clock() - timeNaive_PST
        #------------------------------------------------#
        
#         print "CBA-PST:"
        #-------------------CBA-PST----------------------#
        schedCBA_PST = False
        timeCBA_PST = time.clock() 
        set_mem_Naive(dp)
        memCBA_PST = not mem_overflow_C(ts)
        if memCBA_PST:
            ts.coreAssignPST(0, 4)
            dramCBA_PST = mem_ps_AB_CBA_PST(ts)
        
            memCBA_PST = dramCBA_PST != None
            if memCBA_PST:
                schedCBA_PST = True
                for g_i, tsm in enumerate(ts.modes):
                    psCBA_PST, filler = LPLLCWays(tsm, 'Integer', csch = dramCBA_PST, olap = True, debug = False)
                    if psCBA_PST == None:
                        psCBA_PST, filler = LPLLCWays(tsm, 'Integer', csch = dramCBA_PST, olap = False, debug = False)
                    if psCBA_PST == None:
                        schedCBA_PST = False
                        break
        
        timeCBA_PST = time.clock() - timeCBA_PST
        #------------------------------------------------#
        
        #-------------------CBA-PST-OPT---------------------#
        schedCBA_PST_OPT = False
        timeCBA_PST_OPT = time.clock()
#         for ab_alloc in [4, 8, 12]:
#             ts.coreAssignPST(ab_alloc, 1)
#         
#             memCBA_PST_OPT = not mem_overflow_C(ts)
#             
#             if memCBA_PST_OPT:
#                 psCBA_PST_OPT, filler = LP_CBA_OPT(ts, True, ab_alloc, True)
#                 schedCBA_PST_OPT = psCBA_PST_OPT != None 
#                 if schedCBA_PST_OPT:
#                     break
#         timeCBA_PST_OPT = time.clock() - timeCBA_PST_OPT
        #------------------------------------------------#
    
    
#         print "WBA-PST:"
        #-------------------WBA-PST---------------------#
        schedWBA_PST = False
        timeWBA_PST = time.clock()
        ts.coreAssignPST(0, 16)
    
        set_mem_WBA(dp)
        
        memWBA_PST = not (mem_overflow_AB(ts) or mem_overflow_C(ts))
        
        ts.setLLCAlloc(bypass_scheme())
        
        if memWBA_PST:
            schedWBA_PST = True
            for tsm in ts.modes:
                psWBA_PST, filler = LP_WBA2(tsm, True)
                if psWBA_PST == None:
                    schedWBA_PST = False
                    break
        timeWBA_PST = time.clock() - timeWBA_PST
        #------------------------------------------------#
        
        
        
        ts.remove_buffer_copy_times()
        ts_bef_rep = ts.duplicate()
        ts.replicateShared()      
        ts.add_buffer_copy_times()      
        
        
        
#         print "Naive-RST:"
        #-------------------Naive-RST----------------------#
        schedNaive_RST = False
        timeNaive_RST = time.clock()
        ts.coreAssign(0, 4)
    
        set_mem_Naive(dp)        
        memNaive_RST = not (mem_overflow_AB(ts) or mem_overflow_C(ts))
    
        ts.setLLCAlloc(bypass_scheme())
        
        if memNaive_RST:
            schedNaive_RST = True
            for tsm in ts.modes:
                psNaive_RST, filler = LPLLCWays(tsm, 'Integer', csch = default_scheme(), olap = True, debug = False)
                if psNaive_RST == None:
                    psNaive_RST, filler = LPLLCWays(tsm, 'Integer', csch = default_scheme(), olap = False, debug = False)
                if psNaive_RST == None:
                    schedNaive_RST = False
                    break
                    
        timeNaive_RST = time.clock() - timeNaive_RST
#         print "CBA-RST:"
        #-------------------CBA-RST----------------------#
        schedCBA_RST = False
        timeCBA_RST = time.clock()
        set_mem_Naive(dp)
        memCBA_RST = not mem_overflow_C(ts)
        if memCBA_RST:
            ts.coreAssign(0, 4)
            dramCBA_RST = mem_ps_AB_CBA_RST(ts)
        
            memCBA_RST = dramCBA_RST != None
            if memCBA_RST:
                schedCBA_RST = True
                for g_i, tsm in enumerate(ts.modes):
                    psCBA_RST, filler = LPLLCWays(tsm, 'Integer', csch = dramCBA_RST, olap = True, debug = False)
                    if psCBA_RST == None:
                        psCBA_RST, filler = LPLLCWays(tsm, 'Integer', csch = dramCBA_RST, olap = False, debug = False)
                    if psCBA_RST == None:
                        schedCBA_RST = False
                        break
                        
        timeCBA_RST = time.clock() - timeCBA_RST
        #------------------------------------------------#
        
        #-------------------CBA-RST-OPT---------------------#
        schedCBA_RST_OPT = False
        timeCBA_RST_OPT = time.clock()
#         for ab_alloc in [4, 8, 12]:
#             ts.coreAssign(ab_alloc, 1)
#         
#             memCBA_RST_OPT = not mem_overflow_C(ts)
#         
#             if memCBA_RST_OPT: 
#                 psCBA_RST_OPT, filler = LP_CBA_OPT(ts, False, ab_alloc, True)
#                 schedCBA_RST_OPT = psCBA_RST_OPT != None 
#                 if schedCBA_RST_OPT and memCBA_RST_OPT:
#                     break
# 
#         timeCBA_RST_OPT = time.clock() - timeCBA_RST_OPT
        #------------------------------------------------#
    
    
#         print "WBA-RST:"
        #-------------------WBA-RST--------------------#
        schedWBA_RST = False
        timeWBA_RST = time.clock()
        ts.coreAssign(0, 16)
    
        set_mem_WBA(dp)
        
        memWBA_RST = not (mem_overflow_AB(ts) or mem_overflow_C(ts))
        
        ts.setLLCAlloc(bypass_scheme())
        
        if memWBA_RST:
            schedWBA_RST = True
            for tsm in ts.modes:
                psWBA_RST, filler = LP_WBA2(tsm, False)
                if psWBA_RST == None:
                    schedWBA_RST = False
                    break
                    
        timeWBA_RST = time.clock() - timeWBA_RST
        #------------------------------------------------#
#         print
        ###UEDF###      
        schedEDF = False
        ts.remove_buffer_copy_times()
        ts = ts_bef_rep
                
        set_mem_WBA(dp)
        EDFMemSched = ts.num_pages() < 4*core_pgs_AB + pgs_C
        
        if EDFMemSched:
            cc_reps = [tsk for tsk in ts if tsk.cc_rep]
            for cc_rep in cc_reps:
                ts.remove(cc_rep)
        
            for idx, tsk in enumerate(ts):
                if tsk.level is 'C':
                    break
                for s in range(0, 17):
                    for w in range(0, 17):
                        for crt in {'A', 'B', 'C'}:
                            tsk.cost[crt][s][w] /= loadwtsAB[idx]    
                    
            for idx, tsk in enumerate(ts.level('C')):
                for s in range(0, 17):
                    for w in range(0, 17):
                        for crt in {'A', 'B', 'C'}:
                            tsk.cost[crt][s][w] /= loadwtsC[idx] 
    
            ts.useCiso()
    
            extra_mode_changers = []
            for tsk in ts:
                if tsk.mode_changer:
                    extra_mode_changers.append(tsk)
                else:
                    for s in range(0, 17):
                        for w in range(0, 17):
                            tsk.cost['B'][s][w] = tsk.cost['A'][s][w]   
            for idx, tsk in enumerate(ts.level('C')):
                for s in range(0, 17):
                    for w in range(0, 17):
                        tsk.cost['B'][s][w] /= unmnwts[idx]
    
            
            for tsk in ts:
                tsk.core = 0
                tsk.level = 'B'
    
            ts.setLLCAlloc(uni_scheme())
            ohc.SetOheads(ts, 'pedf')
            mx_wss_all = max([tsk.wss for tsk in ts])
            mx_wss = {'B' : [mx_wss_all, mx_wss_all, mx_wss_all, mx_wss_all],
                      'C' : 0}
            ohc.CPMD = 30
        
            schedEDF = True
            for g_i in range(ts.mode_count):
                if ts.mode(g_i).utilization('B', approximate = False, inflated = True, c_taskset = [], minAperiods = [0, 0, 0, 0], mx_wss = mx_wss) > 1:
                    schedEDF = False
                    break
        
#         if schedNaive_RST and not schedCBA_RST
        sched_res['EDF']             = update_mean2(sched_res['EDF']             , n, schedEDF)
        sched_res['N-PST']           = update_mean2(sched_res['N-PST']           , n, schedNaive_PST)
        sched_res['CBA-PST']         = update_mean2(sched_res['CBA-PST']         , n, schedCBA_PST)
        sched_res['CBA-PST-OPT']     = update_mean2(sched_res['CBA-PST-OPT']     , n, schedCBA_PST_OPT)
        sched_res['WBA-PST']         = update_mean2(sched_res['WBA-PST']         , n, schedWBA_PST)
        sched_res['N-RST']           = update_mean2(sched_res['N-RST']           , n, schedNaive_RST)
        sched_res['CBA-RST']         = update_mean2(sched_res['CBA-RST']         , n, schedCBA_RST)
        sched_res['CBA-RST-OPT']     = update_mean2(sched_res['CBA-RST-OPT']     , n, schedCBA_RST_OPT)
        sched_res['WBA-RST']         = update_mean2(sched_res['WBA-RST']         , n, schedWBA_RST)
        
        sched_res['ALL'] = update_mean2(sched_res['ALL'] , n, schedNaive_PST or schedCBA_PST or schedCBA_PST_OPT or schedWBA_PST or schedNaive_RST or schedCBA_RST or schedCBA_RST_OPT or schedWBA_RST)
        
        time_res['N-PST-Time']       = update_mean2(time_res['N-PST-Time']      , n, timeNaive_PST)
        time_res['CBA-PST-Time']     = update_mean2(time_res['CBA-PST-Time']    , n, timeCBA_PST)
        time_res['CBA-PST-OPT-Time'] = update_mean2(time_res['CBA-PST-OPT-Time'], n, timeCBA_PST_OPT)
        time_res['WBA-PST-Time']     = update_mean2(time_res['WBA-PST-Time']    , n, timeWBA_PST)
        time_res['N-RST-Time']       = update_mean2(time_res['N-RST-Time']      , n, timeNaive_RST)
        time_res['CBA-RST-Time']     = update_mean2(time_res['CBA-RST-Time']    , n, timeCBA_RST)
        time_res['CBA-RST-OPT-Time'] = update_mean2(time_res['CBA-RST-OPT-Time'], n, timeCBA_RST_OPT)
        time_res['WBA-RST-Time']     = update_mean2(time_res['WBA-RST-Time']    , n, timeWBA_RST)
        
        count['Count'] = n+1

        n+=1
        
#         print n, dp.items(), dp.sys_util, str(timeNaive_PST) + "," + str(timeCBA_PST) + "," + str(timeCBA_PST_OPT) + "," + str(timeWBA_PST) + "," + str(timeNaive_RST) + "," + str(timeCBA_RST) + "," + str(timeCBA_RST_OPT) + "," + str(timeWBA_RST)
#         if n%20 == 0:
#             print "checkpoint", n, dp['sys_util']
    
    return dict(dp.items() + sched_res.items() + time_res.items() + count.items())

def myrange(start, end, inc):
    return np.arange(start, end+inc, inc)
    
def valid(dp):
    return True
    
def thread_tests(args, outs, design_points):
    pool = Pool(processes = args.processors)
    try:
        for i, row in enumerate(pool.imap_unordered(process_dp, design_points)):
            if not row:
                continue
#             sys.stderr.write('\rdone {0:%}'.format(i/len(design_points)))
            print 'done {0:%}'.format(i/len(design_points))
#             print '\rdone {0:%}'.format(i/len(design_points))
#             print row
            title = titlefor(row)
            outs[title].writerow(row)
        pool.close()
    except Exception as e:
        print "hit an exception"
        print e
        pool.terminate()
        pool.join()

def titlefor(headers):
    return headers['util_dist'] + "--" + headers['period'] + "--" + headers['task_util'] + "--" + headers['memory_util'] + "--" + headers['dram'] + "--" + headers['pet_sens'] + "--" + headers['mode_count'] + "--" + headers['mode_olap'] + "--" + headers['state_size']

def main():
    parser = argparse.ArgumentParser()
    # parser.add_argument('-d', "--datafile", type = argparse.FileType('w'),
                        # default = sys.stdout,
                        # help = "File to output csv file to")
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help = "Number of processors to execute on")

    args = parser.parse_args()
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.sys_util = myrange(0.2, 9.0, 0.4)
    
    #mode stuff
    exp.dram = NAMED_MEM_FOOTPRINTS.keys()
    
    exp.state_size = NAMED_STATE_SIZE.keys()
    exp.mode_count = NAMED_MODE_COUNT.keys()
    exp.mode_olap = NAMED_MODE_OVERLAP.keys()
    exp.crit_change = NAMED_CRIT_CHANGE.keys()
    exp.util_var = NAMED_UTIL_VAR.keys()
    exp.pet_sens = NAMED_PET_SENS.keys()
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
#     random.shuffle(design_points)
    
    loadohs.load("../oheads/")
    
    # if(len(os.listdir(os.getcwd() + "/results/")) > 0):
        # print "already have results"
        # exit()
#     filenames = ['light-light-heavy--harmonic-long--uni-heavy--uni-heavy--uni-medium--regular--uni-large--uni-light--fixed']
#     filenames = ['light-light-heavy--harmonic-long--uni-heavy--uni-heavy--uni-medium--regular--uni-large--uni-light--fixed',         'medium-medium-medium--harmonic-long--uni-medium--uni-light--uni-heavy--way-heavy--uni-small--uni-heavy--fixed', 'medium-medium-medium--harmonic-long--uni-medium--uni-light--uni-medium--regular--uni-large--uni-light--fixed']
    outs = dict()
    dp_i = 0
    while dp_i < len(design_points):
        dp = design_points[dp_i]
        title = titlefor(dp)
#         if os.path.isfile("../results/" + title):
#         if not title in filenames:
#             design_points.remove(dp)
#             continue
        if not title in outs:
            outs[title] = DictWriter(open("../results_refined/" + title, 'w'), exp.keys()+['EDF', 'N-PST', 'CBA-PST', 'CBA-PST-OPT', 'WBA-PST', 'N-RST', 'CBA-RST', 'CBA-RST-OPT', 'WBA-RST', 'ALL', 'Count', 'N-Time', 'N-PST-Time', 'CBA-PST-Time', 'CBA-PST-OPT-Time', 'WBA-PST-Time', 'N-RST-Time', 'CBA-RST-Time', 'CBA-RST-OPT-Time', 'WBA-RST-Time'])
            outs[title].writeheader()
        dp_i += 1

    print "Total design points: ", len(design_points)
#     for i, t in enumerate(design_points):
#         process_dp(design_points[i])
    start = time.time()
    print "start:", start
    thread_tests(args, outs, design_points)
    end = time.time()
    print ""
    print "end:", end
    
    print "time:", end-start
    print "in hours:", (end-start)/3600.0

if __name__ == '__main__':
    main()